# html_entry.py
""" HTML entry processing functionality """

import html
import logging
import re
import typing
from typing import Optional

import markupsafe

from . import image, links, utils
from .config import config

LOGGER = logging.getLogger(__name__)

HTML_PLAINTEXT_ELEMENTS = (
    'a', 'abbr', 'acronym', 'b', 'bdo', 'cite', 'code', 'dfn', 'em', 'i',
    'kbd', 'label', 'q', 'samp', 'span', 'strong', 'sub', 'sup', 'time', 'tt',
    'var', 'mark', 'p')


class HTMLEntry(utils.HTMLTransform):
    """ An HTML manipulator to fixup src and href attributes """

    def __init__(self, args, search_path):
        super().__init__()

        self._search_path = search_path
        self._config = args

    def handle_decl(self, decl):
        self.append('<!' + decl + '>')

    def handle_data(self, data):
        self.append(data)

    def handle_starttag(self, tag, attrs):
        """ Handle a start tag """
        self._handle_tag(tag, attrs, False)

    def handle_endtag(self, tag):
        """ Handle an end tag """
        self.append('</' + tag + '>')

    def handle_pi(self, data):
        LOGGER.warning("handle_pi: '%s'", data)

    def handle_comment(self, data):
        self.append('<!--' + data + '-->')

    def handle_startendtag(self, tag, attrs):
        """ Handle a self-closing tag """
        self._handle_tag(tag, attrs, True)

    def _handle_tag(self, tag, attrs, self_closing):
        """ Handle a tag.

        attrs -- the attributes of the tag
        self_closing -- whether this is self-closing
        """

        if tag.lower() == 'img':
            attrs = self._image_attrs(attrs)

        # Remap the attributes
        out_attrs = []
        for key, val in attrs:
            if (key.lower() == 'href'
                    or key.startswith('$')
                    or (key.lower() == 'src' and tag.lower() != 'img')):
                if key.startswith('$'):
                    key = key[1:]
                out_attrs.append((key, links.resolve(
                    val, self._search_path, self._config.get('absolute'))))
            else:
                out_attrs.append((key, val))

        self.append(
            utils.make_tag(
                tag,
                out_attrs,
                self_closing))

    def _image_attrs(self, attrs):
        """ Rewrite the SRC attribute on an <img> tag, possibly adding a SRCSET.
        """

        path = None
        attr_args = {}
        is_rewritten = False

        for key, val in attrs:
            if key.lower() == 'width' or key.lower() == 'height':
                try:
                    attr_args[key.lower()] = int(val)
                except ValueError:
                    pass
            elif key.lower() == 'src':
                path = val
            elif key == 'data-publ-rewritten':
                is_rewritten = True
                break

        if is_rewritten:
            # This img tag was already rewritten by a previous processor, so just
            # remove that attribute and return the tag's original attributes
            LOGGER.debug("Detected already-rewritten image; %s", attrs)
            return [(key, val) for key, val in attrs if key != 'data-publ-rewritten']

        if not path:
            # this img doesn't have a src attribute, so there's something unconventional going on
            return attrs

        img_path, img_args, _ = image.parse_image_spec(path)
        img = image.get_image(img_path, self._search_path)

        try:
            img_attrs = img.get_img_attrs({**self._config,
                                           **attr_args,
                                           **img_args})
        except FileNotFoundError as error:
            img_attrs = {
                'data-publ-error': f'File Not Found: {error.filename}'
            }
        except Exception as error:  # pylint:disable=broad-except
            LOGGER.exception("Got exception: %s", error)
            img_attrs = {
                'data-publ-error': f'Error: {error}'
            }

        # return the original attr list with the computed overrides in place
        return [(key, val) for key, val in attrs
                if key.lower() not in img_attrs] + list(img_attrs.items())


def process(text, args, search_path):
    """ Process an HTML entry's HTML """
    args = {**config.layout, **args}
    processor = HTMLEntry(args, search_path)
    processor.feed(text)
    text = processor.get_data()

    if not args.get('markup', True):
        text = strip_html(text)

    return markupsafe.Markup(text)


class HTMLStripper(utils.HTMLTransform):
    """ Strip all HTML tags from a document, except those which are allowed """

    def __init__(self,
                 allowed_tags: Optional[typing.Iterable[str]] = None,
                 allowed_attrs: Optional[typing.Iterable[str]] = None,
                 remove_elements: Optional[typing.Iterable[str]] = None):
        super().__init__()
        self._allowed_tags = set(utils.as_list(allowed_tags))
        self._allowed_attrs = set(utils.as_list(allowed_attrs))
        self._remove_elements = set(utils.as_list(remove_elements))
        self._remove_depth = 0

    def _filter(self, tag, attrs, **kwargs) -> str:
        if self._allowed_tags and tag in self._allowed_tags:
            return utils.make_tag(tag,
                                  {key: val
                                   for key, val in attrs
                                   if key in self._allowed_attrs},
                                  **kwargs)
        if self._remove_elements and tag in self._remove_elements:
            self._remove_depth += 1
        return ''

    def handle_starttag(self, tag, attrs):
        self.append(self._filter(tag, attrs))

    def handle_endtag(self, tag):
        if self._allowed_tags and tag in self._allowed_tags:
            self.append(f'</{tag}>')
        elif self._remove_elements and tag in self._remove_elements:
            self._remove_depth -= 1

    def handle_startendtag(self, tag, attrs):
        self.append(self._filter(tag, attrs, start_end=True))

    def handle_data(self, data):
        if not self._remove_depth:
            self.append(data)


def strip_html(text,
               allowed_tags: Optional[typing.Iterable[str]] = None,
               allowed_attrs: Optional[typing.Iterable[str]] = None,
               remove_elements: Optional[typing.Iterable[str]] = None) -> str:
    """ Strip all HTML formatting off of a chunk of text """
    strip = HTMLStripper(allowed_tags, allowed_attrs, remove_elements)
    strip.feed(str(text))
    return re.sub(r' +', r' ', html.unescape(strip.get_data())).strip()


class FirstParagraph(utils.HTMLTransform):
    """ Get just the first paragraph out of an HTML document """

    def __init__(self, strip_tag=False):
        super().__init__()
        self._strip_tag = strip_tag

        self._found = False  # has text been consumed?
        self._done = False  # have we finished a paragraph?
        self._tag_stack = []  # tuple of tag, consume data, close tag

    @property
    def consuming(self):
        """ Returns whether we're currently consuming data """
        if self._done:
            return False
        if self._tag_stack:
            return self._tag_stack[-1][1]
        return True

    def handle_starttag(self, tag, attrs):
        if tag.lower() == 'p':
            # we're done if we've already encountered text outside of a paragraph
            self._done = self._found
            consume = not self._done
        else:
            consume = self.consuming and tag.lower() in HTML_PLAINTEXT_ELEMENTS

        insert_tag = consume and (self._tag_stack or not self._strip_tag)
        if insert_tag:
            self.append(utils.make_tag(tag, attrs))

        self._tag_stack.append((tag.lower(), consume, insert_tag))

    def handle_endtag(self, tag):
        popped = None
        inserted = False

        # close tags until we've closed the current one
        while self._tag_stack and popped != tag.lower():
            popped, _, inserted = self._tag_stack.pop()
            if inserted:
                self.append(f'</{popped}>')

        # we're done if we've closed a paragraph and text has been consumed
        if tag.lower() == 'p' and self._found:
            self._done = True

    def handle_startendtag(self, tag, attrs):
        if self.consuming and tag in HTML_PLAINTEXT_ELEMENTS:
            self.append(utils.make_tag(tag, attrs, True))

    def handle_data(self, data):
        if self.consuming:
            if data.strip():
                self._found = True
            self.append(data)


def first_paragraph(text, strip_tag=False):
    """ Extract the first paragraph of text from an HTML document """
    first_para = FirstParagraph(strip_tag=strip_tag)
    first_para.feed(str(text))
    return markupsafe.Markup(first_para.get_data())



# view.py
""" A view of entries """

import logging
import typing

import arrow
import flask
from werkzeug.utils import cached_property

from . import caching
from . import category as publ_category
from . import queries, tokens, user, utils
from .entry import Entry

ViewSpec = typing.Dict[str, typing.Any]  # pylint:disable=invalid-name
PaginationType = typing.Union[utils.TimeSpan, typing.Literal['count'], None]

LOGGER = logging.getLogger(__name__)

# Prioritization list for pagination
OFFSET_PRIORITY = ['date', 'start', 'last', 'first']

# Prioritization list for pagination type
#
# NOTE: date appears in here too so that if it appears as an offset it
# overrides the count
PAGINATION_PRIORITY = ['date', 'count']

# All spec keys that indicate a pagination
PAGINATION_SPECS = OFFSET_PRIORITY + PAGINATION_PRIORITY

SPAN_FORMATS = {
    'day': 'YYYY/MM/DD',
    'week': 'YYYY/MM/DD',
    'month': 'YYYY/MM',
    'year': 'YYYY/MM'
}


class View(caching.Memoizable):
    # pylint: disable=too-many-instance-attributes,too-many-public-methods
    """ A view of entries """

    __hash__ = caching.Memoizable.__hash__  # type:ignore

    @staticmethod
    @utils.stash
    def load(input_spec: ViewSpec):
        """ Generate a view.

        input_spec -- the parameters to the view. In addition to the values provided
        to queries.build_query, this also takes the following keys:
            count -- How many entries to include in a page
            order -- How to order the entries ('newest' or 'oldest')
        """
        return View(View.load.__name__, input_spec)

    def __init__(self, create_key, input_spec: ViewSpec):
        """ Instantiate the view """

        assert create_key == View.load.__name__, "View must be created with View.load()"

        # filter out any priority override things
        spec: ViewSpec = {
            k: v for k, v in input_spec.items()
            if k not in PAGINATION_SPECS
        }

        # pull in the first offset type that appears
        for offset in OFFSET_PRIORITY:
            if offset in input_spec:
                spec[offset] = input_spec[offset]
                break

        # pull in the first page type that appears
        paginated = False
        for pagination in PAGINATION_PRIORITY:
            if pagination in input_spec:
                paginated = True
                spec[pagination] = input_spec[pagination]
                break

        self.spec = spec.copy()

        self._order_by = spec.pop('order', 'newest').lower()

        if 'start' in spec and paginated:
            if self._order_by == 'oldest':
                spec['first'] = spec['start']
            elif self._order_by == 'newest':
                spec['last'] = spec['start']
        spec.pop('start', None)

        count = spec.pop('count', None)

        self._deleted = queries.build_query({
            **spec,
            '_deleted': True
        }).order_by(*queries.ORDER_BY[self._order_by])
        self._entries = queries.build_query(
            spec).order_by(*queries.ORDER_BY[self._order_by])

        self.type: PaginationType = None
        if self.spec.get('date') is not None:
            _, self.type, _ = utils.parse_date(self.spec['date'])
        elif count is not None:
            self.type = 'count'

    def _key(self):
        return repr(self.spec)

    def __str__(self):
        return str(self.link())

    @cached_property
    def first(self) -> typing.Optional[Entry]:
        """ Gets the first entry in the view """
        entries = self.entries()
        return entries[0] if entries else None

    @cached_property
    def last(self) -> typing.Optional[Entry]:
        """ Gets the last entry in the view """
        entries = self.entries()
        return entries[-1] if entries else None

    @cached_property
    def entries(self) -> typing.Callable[..., typing.List[Entry]]:
        """ Gets entries which are authorized for the current viewer """

        def _entries(unauthorized=0) -> typing.List[Entry]:
            count = self.spec.get('count')
            return Entry.filter_auth(self._entries, count, unauthorized)

        return utils.CallableProxy(_entries)

    @cached_property
    def has_unauthorized(self) -> bool:
        """ Returns whether there's entries that the user could see if they
        were authorized differently """

        LOGGER.debug("HAS_UNAUTHORIZED")

        count = self.spec.get('count')
        cur_user = user.get_active()

        if cur_user and cur_user.is_admin:
            return False

        auth_count = 0

        LOGGER.debug("Checking for unauthorized in the first %s entries", count)

        # retrieve entries a chunk at a time so Pony doesn't have to realize
        # the whole table up-front
        start = 0
        chunk_size = max(count if count is not None else 0, 16)
        while count is None or auth_count < count:
            LOGGER.debug("Testing %d", start)
            chunk = self._entries[start:start + chunk_size]
            start += chunk_size

            if not chunk:
                LOGGER.debug("Hit EOF")
                break

            for record in chunk:
                if not record.is_authorized(cur_user):
                    LOGGER.debug("Found unauthorized entry")
                    tokens.request(cur_user)
                    return True
                auth_count += 1
                LOGGER.debug("Seen %d authorized", auth_count)

        return False

    @cached_property
    def unauthorized(self) -> typing.Callable[..., typing.List[Entry]]:
        """ Gets the entries which would be visible on the view if the user
        were authorized to see them """

        def _unauthorized(count=None) -> typing.List[Entry]:
            result: typing.List[Entry] = []

            auth_max = self.spec.get('count')
            cur_user = user.get_active()

            if cur_user and cur_user.is_admin:
                return result

            auth_count = 0
            start = 0
            chunk_size = max(auth_max if auth_max is not None else 0, 16)
            while ((auth_max is None or auth_count < auth_max)
                   and (count is None or len(result) < count)):

                chunk = self._entries[start:start + chunk_size]
                start += chunk_size

                if not chunk:
                    break

                for record in chunk:
                    # if we've traversed the whole authorized count, exit
                    if auth_max is not None and auth_count >= auth_max:
                        break

                    # if we've reached our unauthorized max, exit
                    if count is not None and len(result) >= count:
                        break

                    if not record.is_authorized(cur_user):
                        tokens.request(cur_user)
                        result.append(Entry.load(record))
                    else:
                        auth_count += 1

            return result

        return utils.CallableProxy(_unauthorized)

    @cached_property
    def deleted(self) -> typing.List[Entry]:
        """ Gets the deleted entries from the view """

        query = self._deleted

        if self.spec.get('count'):
            # We're using a count constraint, so we'll want to limit this to
            # entries that come before our next page of non-deleted results.
            #
            # We only use the next-page start because our pagination constraint
            # (if any) will have already limited this query appropriately, and
            # we consider deleted entries in the gap between two views to
            # belong to the first page of views.
            if self._order_by == 'oldest' and self.next:
                query = queries.where_before_entry(query, self.next.first)
            elif self._order_by == 'newest' and self.next:
                query = queries.where_after_entry(query, self.next.first)

        return [Entry.load(e) for e in query]

    @cached_property
    def count(self) -> int:
        """ Returns the number of entries in the view """
        return len(self.entries())

    @cached_property
    def last_modified(self) -> arrow.Arrow:
        """ Gets the most recent modification time for all entries in the view """
        entries = self.entries()
        if entries:
            latest = max(entries, key=lambda x: x.last_modified)
            return arrow.get(latest.last_modified)
        return arrow.get()

    @cached_property
    def older(self) -> typing.Optional['View']:
        """ Gets the page of older items """
        older, _ = self._pagination
        return older

    @cached_property
    def newer(self) -> typing.Optional['View']:
        """ Gets the page of newer items """
        _, newer = self._pagination
        return newer

    @cached_property
    def previous(self) -> typing.Optional['View']:
        """ Gets the previous page, respecting sort order """
        if self._order_by == 'oldest':
            return self.older
        if self._order_by == 'newest':
            return self.newer
        return None

    @cached_property
    def next(self) -> typing.Optional['View']:
        """ Gets the next page, respecting sort order """
        if self._order_by == 'oldest':
            return self.newer
        if self._order_by == 'newest':
            return self.older
        return None

    @cached_property
    def newest(self) -> typing.Optional[Entry]:
        """ Gets the newest entry in the view, regardless of sort order """
        if self._order_by == 'newest':
            return self.first
        if self._order_by == 'oldest':
            return self.last
        return max(self.entries(), key=lambda x: (x.date, x.id))

    @cached_property
    def oldest(self) -> typing.Optional[Entry]:
        """ Gets the oldest entry in the view, regardless of sort order """
        if self._order_by == 'newest':
            return self.last
        if self._order_by == 'oldest':
            return self.first
        return min(self.entries(), key=lambda x: (x.date, -x.id))

    @cached_property
    def paging(self) -> str:
        """ Gets the pagination type; compatible with entry.archive(page_type=...) """
        if 'date' in self.spec:
            _, date_span, _ = utils.parse_date(self.spec['date'])
            return date_span
        return 'offset'

    @cached_property
    def pages(self) -> typing.List['View']:
        """ Gets a list of all pages for this view """
        pages = []

        cur: typing.Optional[View] = self.previous
        while cur is not None:
            pages.append(cur)
            cur = cur.previous
        pages.reverse()

        cur = self
        while cur is not None:
            pages.append(cur)
            cur = cur.next

        return pages

    @cached_property
    def range(self) -> typing.Callable[..., str]:
        """ Gets a localizable string describing the view range """
        def _view_name(**formats) -> str:
            if not any(k for k in PAGINATION_SPECS if k in self.spec):
                # We don't have anything that specifies a pagination constraint, so
                # we don't have a name
                return ''

            if not self.oldest or not self.newest:
                # We don't have any entries, so we don't have a name
                return ''

            if 'date' in self.spec:
                _, span_type, span_format = utils.parse_date(self.spec['date'])
            elif self.oldest.date.year != self.newest.date.year:
                span_type = 'year'
                span_format = utils.YEAR_FORMAT
            elif self.oldest.date.month != self.newest.date.month:
                span_type = 'month'
                span_format = utils.MONTH_FORMAT
            else:
                span_type = 'day'
                span_format = utils.DAY_FORMAT

            date_format = formats.get(
                span_type, SPAN_FORMATS.get(span_type, span_format))

            oldest = self.oldest.date.format(date_format)
            if self.oldest == self.newest:
                return oldest

            newest = self.newest.date.format(date_format)

            if oldest == newest:
                template = formats.get('single', '{oldest} ({count})')
            else:
                template = formats.get(
                    'span', '{oldest} — {newest} ({count})')

            return template.format(count=len(self.entries()),
                                   oldest=oldest,
                                   newest=newest)
        return utils.CallableProxy(_view_name)

    @cached_property
    def link(self) -> typing.Callable[..., str]:
        """ Gets a link back to this view """

        def _link(template='', absolute=False, category=None, **kwargs) -> str:
            args = {}
            if 'date' in self.spec:
                args['date'] = self.spec['date']
            else:
                for k in OFFSET_PRIORITY:
                    if k in self.spec:
                        val = self.spec[k]
                        if isinstance(val, (str, int)):
                            args['id'] = val
                        elif hasattr(val, 'start'):
                            # the item was an object, so we want the object's
                            # id
                            args['id'] = val.id
                        else:
                            raise ValueError(f"key {k} is of type {type(val)}")
                        break

            taglist = kwargs.get('tag', self.spec.get('tag'))
            if taglist:
                args['tag'] = list(utils.TagSet(utils.as_list(taglist)).keys())

            view_category = publ_category.Category.load(category if category
                                                        else self.spec.get('category'))

            return flask.url_for('category',
                                 **args,
                                 template=template if str(template) != view_category.index_template
                                 else '',
                                 category=category if category
                                 else self.spec.get('category'),
                                 _external=absolute,
                                 **kwargs)

        return utils.CallableProxy(_link)

    @cached_property
    def tags(self) -> utils.TagSet:
        """ Returns a list of all the tags applied to this view """
        tag_list = self.spec.get('tag', [])
        return utils.TagSet(utils.as_list(tag_list))

    @cached_property
    def current(self) -> typing.Callable[..., 'View']:
        """ Gets a version of this view without any pagination offsets """

        def _get_current(**restrict) -> 'View':
            spec = {k: v for (k, v) in self.spec.items()
                    if k not in OFFSET_PRIORITY}
            return View.load({**spec, **restrict})

        return utils.CallableProxy(_get_current)

    @cached_property
    def is_current(self) -> bool:
        """ Returns true if this is equivalent to self.current """
        for k in self.spec.keys():
            if k in OFFSET_PRIORITY:
                return False
        return True

    @cached_property
    def _pagination(self) -> typing.Tuple[typing.Optional['View'], typing.Optional['View']]:
        """ Compute the neighboring pages from this view.

        Returns a tuple of older page, newer page.
        """

        base = {key: val for key, val in self.spec.items()
                if key not in OFFSET_PRIORITY}

        if 'date' in self.spec:
            LOGGER.debug('by date')
            return self._get_date_pagination(base)

        if 'count' in self.spec:
            LOGGER.debug('by count')
            return self._get_count_pagination(base)

        # we're not paginating
        return None, None

    @property
    def query_spec(self):
        """ Get the view spec in a queryable form """
        return self.filter_query_spec(self.spec)

    @staticmethod
    def filter_query_spec(spec: ViewSpec):
        """ Filter out query parameters that are view-specific """
        return {k: v for k, v in spec.items() if k not in ('order', 'count')}

    def _get_date_pagination(self,
                             base: ViewSpec
                             ) -> typing.Tuple[typing.Optional['View'], typing.Optional['View']]:
        """ Compute the pagination for date-based views """
        date, interval, date_format = utils.parse_date(self.spec['date'])
        start_date, end_date = date.span(interval)

        base_query = queries.build_query(self.filter_query_spec(base))

        oldest_neighbor = base_query.filter(lambda e: e.local_date < start_date.datetime)\
            .order_by(*queries.ORDER_BY['newest']).first()
        newest_neighbor = base_query.filter(lambda e: e.local_date > end_date.datetime)\
            .order_by(*queries.ORDER_BY['oldest']).first()

        older_view: typing.Optional['View'] = None
        newer_view: typing.Optional['View'] = None

        if newest_neighbor:
            newer_date = newest_neighbor.display_date
            newer_view = View.load({**base,
                                    'date': arrow.get(newer_date).format(date_format)})

        if oldest_neighbor:
            older_date = oldest_neighbor.display_date
            older_view = View.load({**base,
                                    'date': arrow.get(older_date).format(date_format)})

        return older_view, newer_view

    def _get_count_pagination(self,
                              base: ViewSpec
                              ) -> typing.Tuple[typing.Optional['View'], typing.Optional['View']]:
        """ Compute the pagination for count-based views """

        oldest = self.oldest
        newest = self.newest

        oldest_neighbor = View.load({
            **base,
            'before': oldest,
            'order': 'newest'
        }).first if oldest else None

        newest_neighbor = View.load({
            **base,
            'after': newest,
            'order': 'oldest'
        }).first if newest else None

        count = self.spec['count']

        out_spec = {**base, 'count': count, 'order': self._order_by}

        if self._order_by == 'newest':
            older_view = View.load({**out_spec,
                                    'last': oldest_neighbor}) if oldest_neighbor else None

            newer_count = View.load({**base,
                                     'first': newest_neighbor,
                                     'order': 'oldest',
                                     'count': count}) if newest_neighbor else None
            newer_view = View.load({**out_spec,
                                    'last': newer_count.last}) if newer_count else None

            return older_view, newer_view

        if self._order_by == 'oldest':
            older_count = View.load({**base,
                                     'last': oldest_neighbor,
                                     'order': 'newest',
                                     'count': count}) if oldest_neighbor else None
            older_view = View.load({**out_spec,
                                    'first': older_count.last}) if older_count else None

            newer_view = View.load({**out_spec,
                                    'first': newest_neighbor}) if newest_neighbor else None

            return older_view, newer_view

        return None, None

    def __call__(self, **restrict):
        return View.load({**self.spec, **restrict})

    def __iter__(self):
        return self.entries().__iter__()

    def tag_add(self, *tags: utils.ListLike[str]) -> 'View':
        """ Return a view with the specified tags added """
        return View.load({**self.spec, 'tag': self.tags | set(tags)})

    def tag_remove(self, *tags: utils.ListLike[str]) -> 'View':
        """ Return a view with the specified tags removed """
        return View.load({**self.spec, 'tag': self.tags - set(tags)})

    def tag_toggle(self, *tags: utils.ListLike[str]) -> 'View':
        """ Return a view with the specified tags toggled """
        return View.load({**self.spec, 'tag': self.tags ^ set(tags)})


def get_view(**kwargs) -> View:
    """ Wrapper function for constructing a view from scratch """
    return View.load(input_spec=kwargs)


def parse_view_spec(args) -> ViewSpec:
    """ Parse a view specification from a request arg list """

    view_spec = {}

    if 'date' in args:
        view_spec['date'] = args['date']
    elif 'id' in args:
        view_spec['start'] = args['id']

    if 'tag' in args:
        view_spec['tag'] = args.getlist('tag')
        if len(view_spec['tag']) == 1:
            view_spec['tag'] = args['tag']

    return view_spec



""" IndieAuth token endpoint """

import functools
import json
import logging
import time
import typing
import urllib.parse
from typing import Optional

import arrow
import bs4
import flask
import itsdangerous
import requests
import werkzeug.exceptions as http_error
from pony import orm

from . import model, utils
from .config import config

LOGGER = logging.getLogger(__name__)


def signer(context: str):
    """ Gets the signer/validator for the tokens """
    from .flask_wrapper import current_app
    return itsdangerous.URLSafeSerializer(str(current_app.secret_key) + context)


def get_token(id_url: str, lifetime: int, scope: Optional[str] = None, context: str = '') -> str:
    """ Gets a signed token for the given identity"""
    token = {'me': utils.canonicize_url(id_url)}
    if scope:
        token['scope'] = scope

    return signer(context).dumps((token, int(time.time() + lifetime)))


def parse_token(token: str, context: str = '') -> typing.Dict[str, str]:
    """ Parse a bearer token to get the stored data """
    try:
        ident, expires = signer(context).loads(token)
    except itsdangerous.BadData as error:
        LOGGER.error("Got token parse error: %s", error)
        flask.g.token_error = 'Invalid token'  # pylint:disable=assigning-non-slot
        raise http_error.Unauthorized('Invalid token') from error

    if expires < time.time():
        LOGGER.info("Got expired token for %s", ident['me'])
        flask.g.token_error = "Token expired"  # pylint:disable=assigning-non-slot
        raise http_error.Unauthorized("Token expired")

    return ident


def request(user):
    """ Called whenever an authenticated access fails; marks authentication
    as being upgradeable.

    Currently this is unused by Publ itself, but a site can make use of it to
    e.g. add a ``WWW-Authenticate`` header or the like in a post-request hook.
    """

    if not user:
        flask.g.needs_auth = True  # pylint:disable=assigning-non-slot


def send_auth_ticket(subject: str,
                     resource: str,
                     endpoint: str,
                     scope: Optional[str] = None):
    """ Initiate the TicketAuth flow """
    from .flask_wrapper import current_app

    def _submit():
        ticket = get_token(subject, config.ticket_lifetime, scope, context='ticket')

        req = requests.post(endpoint, data={
            'ticket': ticket,
            'resource': resource,
            'subject': subject
        }, timeout=30)
        LOGGER.info("Auth ticket sent to %s for %s: %d %s",
                    endpoint, subject, req.status_code, req.text)

    # Use the indexer's threadpool to issue the ticket in the background
    current_app.indexer.submit(_submit)


@orm.db_session()
def log_grant(identity: str):
    """ Update the user table with the granted token """
    import authl.handlers.indieauth

    values = {
        'last_token': arrow.utcnow().datetime,
    }

    profile = authl.handlers.indieauth.get_profile(identity)
    if profile:
        values['profile'] = profile

    record = model.KnownUser.get(user=identity)
    if record:
        record.set(**values)
    else:
        record = model.KnownUser(user=identity,
                                 **values,
                                 last_seen=arrow.utcnow().datetime)


def redeem_grant(grant_type: str, auth_token: str):
    """ Redeem a grant from a provided redemption ticket """
    grant = parse_token(auth_token, grant_type)
    LOGGER.info("Redeeming %s for %s; scopes=%s", grant_type, grant['me'],
                grant.get('scope'))

    scope = grant.get('scope', '')

    token = get_token(grant['me'], config.token_lifetime, scope)

    response = {
        'access_token': token,
        'token_type': 'Bearer',
        'me': grant['me'],
        'expires_in': config.token_lifetime,
        'refresh_token': get_token(grant['me'],
                                   config.refresh_token_lifetime,
                                   scope,
                                   context='refresh_token')
    }
    if scope:
        response['scope'] = scope

    log_grant(grant['me'])

    return json.dumps(response), {'Content-Type': 'application/json'}


@functools.lru_cache()
def get_ticket_endpoint(me_url: str):
    """ Get the IndieAuth Ticket Auth endpoint and the canonical identity URL """
    LOGGER.debug("get_ticket_endpoint %s", me_url)
    import authl.handlers.indieauth
    from bs4 import BeautifulSoup

    req = authl.utils.request_url(me_url)
    content = BeautifulSoup(req.text, 'html.parser')

    if req.links and 'canonical' in req.links:
        canonical_url = req.links['canonical']['url']
    else:
        link = typing.cast(bs4.Tag, content.find('link', rel='canonical'))
        if link:
            canonical_url = urllib.parse.urljoin(me_url,
                                                 typing.cast(str, link.get('href')))
        else:
            canonical_url = me_url

    if utils.canonicize_url(canonical_url) != utils.canonicize_url(me_url):
        # We have a rel="canonical" which mismatches the provided identity URL
        LOGGER.debug("%s -> canonical=%s", me_url, canonical_url)
        endpoint, me_url = authl.handlers.indieauth.find_endpoint(canonical_url,
                                                                  rel='ticket_endpoint')
    else:
        # Use our fetch to seed Authl's endpoint fetch and get that instead
        endpoints, me_url = authl.handlers.indieauth.find_endpoints(me_url,
                                                                    req.links, content)
        endpoint = endpoints.get('ticket_endpoint')

    LOGGER.debug("%s %s", me_url, endpoint)
    return endpoint, me_url


def ticket_request(me_url: str, scope: str):
    """ Initiate a ticket request """

    try:
        endpoint, me_url = get_ticket_endpoint(utils.canonicize_url(me_url))
    except RuntimeError:
        endpoint = None
    if not endpoint:
        raise http_error.BadRequest("Could not get ticket endpoint")
    LOGGER.info("endpoint: %s", endpoint)
    send_auth_ticket(me_url, flask.request.url_root, endpoint, scope)
    return "Ticket sent", 202


def parse_authorization_header(header):
    """ Parse an Authorization: header from an HTTP request into token """
    parts = header.split()
    if len(parts) < 2:
        LOGGER.warning("Got malformed authorization header '%s'", header)
        raise http_error.BadRequest("Malformed authorization header")

    if parts[0].lower() == 'bearer':
        token = parse_token(parts[1])
        return token

    raise http_error.BadRequest(f"Unknown authorization type '{parts[0]}'")


def indieauth_endpoint():
    """ IndieAuth token endpoint """

    if 'grant_type' in flask.request.form:
        # token grant
        if flask.request.form['grant_type'] == 'ticket':
            # TicketAuth
            if 'ticket' not in flask.request.form:
                raise http_error.BadRequest("Missing ticket")
            return redeem_grant('ticket', flask.request.form['ticket'])

        if flask.request.form['grant_type'] == 'refresh_token':
            # Refresh token redemption
            if 'refresh_token' not in flask.request.form:
                raise http_error.BadRequest("Missing refresh_token")
            return redeem_grant('refresh_token', flask.request.form['refresh_token'])

        raise http_error.BadRequest("Unknown grant type")

    if 'action' in flask.request.form:
        # provisional ticket request flow, per https://github.com/indieweb/indieauth/issues/87
        if flask.request.form['action'] == 'ticket' and 'subject' in flask.request.form:
            return ticket_request(flask.request.form['subject'],
                                  flask.request.form.get('scope', ''))

        raise http_error.BadRequest()

    if 'Authorization' in flask.request.headers:
        # ticket verification
        token = parse_authorization_header(flask.request.headers['Authorization'])
        return json.dumps(token), {'Content-Type': 'application/json'}

    if 'me' in flask.request.args:
        # ad-hoc ticket request
        return ticket_request(flask.request.args['me'],
                              flask.request.args.get('scope', ''))

    raise http_error.BadRequest()



# template.py
""" Wrapper for template information """

import fnmatch
import glob
import hashlib
import logging
import mimetypes
import os
import typing

import arrow
import flask
import werkzeug.exceptions as http_error

from . import image, utils
from .config import config

LOGGER = logging.getLogger(__name__)

BUILTIN_DIR = os.path.join(os.path.dirname(__file__), 'default_template')

# A useful set of default priorities for clients that don't declare Accept (e.g. curl)
DEFAULT_ACCEPT = ['text/html',
                  'application/rss+xml',
                  'application/atom+xml',
                  'application/xml',
                  'style/css',
                  'text/plain',
                  '*/*']


class Template:
    """ Template information wrapper """
    # pylint: disable=too-few-public-methods,too-many-instance-attributes

    def __init__(self, name: str, filename: str, file_path: typing.Optional[str],
                 content: typing.Optional[str] = None,
                 mime_type: typing.Optional[str] = None):
        """ Useful information for the template object:

        name -- The name of the template
        filename -- The filename of the template
        file_path -- The full path to the template
        content -- static content
        """
        # pylint:disable=too-many-arguments,too-many-positional-arguments
        self.name = name

        # Flask expects template filenames to be /-separated regardless of
        # platform
        if os.sep != '/':
            self.filename = '/'.join(os.path.normpath(filename).split(os.sep))
        else:
            self.filename = filename

        self.file_path = file_path

        if file_path:
            self.mtime = os.stat(file_path).st_mtime
            self.last_modified = arrow.get(self.mtime)
            self._fingerprint = utils.file_fingerprint(file_path)

        self.content = content
        if content:
            self._fingerprint = hashlib.md5(content.encode('utf-8')).hexdigest()

        self.mime_type = mime_type if mime_type else mimetypes.guess_type(filename)[0]

    def render(self, **args) -> str:
        """ Render the template with the appropriate Flask function """
        if self.content:
            return flask.render_template_string(self.content, **args)
        return flask.render_template(self.filename, **args)

    def __str__(self):
        return self.name

    def _key(self):
        return Template, self.file_path, self._fingerprint

    def __repr__(self):
        return repr(self._key())

    def __hash__(self):
        return hash(self._key())

    def image(self, filename):
        """ Retrieve an image using our search path as the context """
        search_path = (os.path.join(config.content_folder, os.path.dirname(self.filename)),)
        return image.get_image(filename, search_path)


def map_template(category: str,
                 template_list: typing.Union[str, typing.List[str]],
                 in_exception=False
                 ) -> typing.Optional[Template]:
    """
    Given a file path and an acceptable list of templates, return the
    best-matching template's path relative to the configured template
    directory.

    Arguments:

    category -- The path to map
    template_list -- A template to look up (as a string), or a list of templates.
    """
    # pylint:disable=too-many-locals,too-many-branches

    # get the sorted acceptance list
    accept_mime = [mime for (mime, _) in flask.request.accept_mimetypes]
    if in_exception:
        accept_mime.append('*/*')

    # Clients which accept *anything* (e.g. curl) should be forced into a more
    # sensible priority order
    if not accept_mime or accept_mime == ['*/*']:
        accept_mime = DEFAULT_ACCEPT

    # Get the MIME types that are also just glob matches
    accept_glob = [mime for mime in accept_mime if '*' in mime]

    LOGGER.debug("accept_mime: %s", accept_mime)
    LOGGER.debug("accep_glob: %s", accept_glob)

    # gets a list like [('foo/bar', ['.foo','.bar'])]
    mime_exts = [(None, [''])] + [(mime, mimetypes.guess_all_extensions(mime))
                                  for mime in accept_mime]

    # unpacks it to [('foo/bar', '.foo'), ('foo/bar', '.bar')]
    extensions = [(mime, ext) for mime, exts in mime_exts for ext in exts]

    LOGGER.debug("Extension priority: %s", extensions)

    could_glob = False

    for template in utils.as_list(template_list):
        path: typing.Optional[str] = os.path.normpath(category)
        while path is not None:
            LOGGER.debug("checking path %s for template %s", path, template)
            for mime, extension in extensions:
                LOGGER.debug('path=%s mime=%s extension=%s', path, mime, extension)
                candidate = os.path.join(path, template + extension)
                file_path = os.path.join(config.template_folder, candidate)
                if os.path.isfile(file_path):
                    # Note that if the template is called out directly, this will
                    # not check if it matches the Accept: header. This technically
                    # violates HTTP but in any situation where the name is given
                    # directly there's almost certainly a */* in place.
                    #
                    # Properly checking Accept: in this context would be super annoying.
                    return Template(template, candidate, file_path,
                                    mime_type=mime)

            # check for glob matches
            glob_candidate = os.path.join(path, f'{template}.*')
            glob_files = glob.glob(glob_candidate, root_dir=config.template_folder)
            if glob_files:
                could_glob = True

            for pattern in accept_glob:
                for candidate in glob_files:
                    cmime, _ = mimetypes.guess_type(candidate)
                    if pattern == '*/*' or (cmime and fnmatch.fnmatch(cmime, pattern)):
                        LOGGER.debug("Found glob match: %s (%s)", candidate, cmime)
                        return Template(template, candidate,
                                        os.path.join(config.template_folder, candidate))

            parent = os.path.dirname(path)
            if parent != path:
                path = parent
            else:
                path = None

    # We didn't find one in the filesystem, so let's consult the builtins instead
    for template in utils.as_list(template_list):
        for mime, extension in extensions:
            filename = template + extension
            template_string = _get_builtin(filename)
            if template_string:
                return Template(template, filename, None,
                                content=template_string, mime_type=mime)

        # check for glob matches
        glob_files = glob.glob(f'{template}.*', root_dir=BUILTIN_DIR)
        if glob_files:
            could_glob = True

        for pattern in accept_glob:
            for candidate in glob_files:
                cmime, _ = mimetypes.guess_type(candidate)
                if cmime and fnmatch.fnmatch(cmime, pattern):
                    return Template(template, candidate, None, content=_get_builtin(candidate))

    if could_glob:
        # A precise match wasn't found, but could have been if there were a broader acceptance
        raise http_error.NotAcceptable(f"Could not find match for {accept_mime}")

    return None


def _get_builtin(filename: str) -> typing.Optional[str]:
    """ Get a builtin template """

    builtin_file = os.path.join(BUILTIN_DIR, filename)
    if os.path.isfile(builtin_file):
        with open(builtin_file, 'r', encoding='utf-8') as file:
            return file.read()

    return None



""" Full-text search stuff """
import datetime
import email
import logging
import os
import typing

try:
    import whoosh
    import whoosh.fields
    import whoosh.index
    import whoosh.qparser
    import whoosh.query
    import whoosh.writing
except ImportError:
    whoosh = None

from pony import orm
from werkzeug.utils import cached_property

from . import model, tokens, user, utils
from .entry import Entry

LOGGER = logging.getLogger(__name__)
SCHEMA_VERSION = 2


class SearchResults:
    """ Result set from a full-text search """

    def __init__(self, results):
        self._entries = [record
                         for record in [model.Entry.get(id=int(hit['entry_id']))
                                        for hit in results]
                         if record and record.visible]

    @cached_property
    def has_unauthorized(self):
        """ Returns whether there are any unauthorized entries present """

        cur_user = user.get_active()

        for record in self._entries:
            if not record.is_authorized(cur_user):
                tokens.request(cur_user)
                return True
        return False

    @cached_property
    def entries(self) -> typing.Callable[..., typing.List[Entry]]:
        """ Returns the entries in the search results """
        def _entries(unauthorized=0) -> typing.List[Entry]:
            return Entry.filter_auth(self._entries, unauthorized=unauthorized)

        return utils.CallableProxy(_entries)

    def __iter__(self):
        return self.entries().__iter__()

    def __len__(self):
        return len(self._entries)

    def __bool__(self):
        return bool(self._entries)


class SearchIndex:
    """ Full-text search index for all entries """

    def __init__(self, config):
        if not config.search_index:
            self.index = None
            return

        if not whoosh:
            self.index = None
            LOGGER.error(
                "Search index configured but required libraries are not installed. "
                "See https://publ.plaidweb.site/manual/865-Python-API#search_index")

        self.schema = whoosh.fields.Schema(
            entry_id=whoosh.fields.ID(stored=True, unique=True),
            title=whoosh.fields.TEXT,
            content=whoosh.fields.TEXT,
            published=whoosh.fields.DATETIME,
            tag=whoosh.fields.KEYWORD(lowercase=True, commas=True),
            category=whoosh.fields.ID,
            status=whoosh.fields.NUMERIC)

        if not os.path.exists(config.search_index):
            os.mkdir(config.search_index)

        with orm.db_session:
            version = model.GlobalConfig.get(key='search_index_version')

            if (not whoosh.index.exists_in(config.search_index)
                or not version
                    or version.int_value != SCHEMA_VERSION):
                # either the index doesn't exist or it's got the wrong schema;
                # (re)create it
                self.index = whoosh.index.create_in(config.search_index, self.schema)
                if version:
                    version.int_value = SCHEMA_VERSION
                else:
                    version = model.GlobalConfig(key='search_index_version',
                                                 int_value=SCHEMA_VERSION)
            else:
                self.index = whoosh.index.open_dir(config.search_index)

        self.query_parser = whoosh.qparser.QueryParser("content", self.index.schema)

    @property
    def active(self):
        """ Return whether the search index is active """
        return self.index is not None

    def update(self, record: model.Entry,
               entry_file: typing.Optional[email.message.EmailMessage]):
        """
        Add an entry to the content index
        """

        if not self.index:
            return

        if record.status not in (model.PublishStatus.PUBLISHED.value,
                                 model.PublishStatus.SCHEDULED.value):
            self.index.delete_by_term("entry_id", str(record.id))
            return

        with whoosh.writing.AsyncWriter(self.index) as writer:
            writer.update_document(
                entry_id=str(record.id),
                title=record.title,
                content=entry_file.get_payload() if entry_file else '',
                published=datetime.datetime.fromtimestamp(record.utc_timestamp),
                tag=','.join(entry_file.get_all('tag') or []) if entry_file else '',
                category=record.category,
                status=record.status)

    def remove(self, entry_id: int):
        """ Remove an entry by ID """
        if not self.index:
            return
        self.index.delete_by_term("entry_id", str(entry_id))

    def query(self, query: str,
              category=None, recurse=False,
              count=None, page=None,
              future=False):
        """
        Searches with a text query

        :param category: The category to search on; None to avoid restriction
        :param bool recurse: Whether to also match subcategories
        :param int count: The number of entries per page
        :param int page: The page number to retrieve
        :param bool future: Whether to retrieve entries that aren't yet visible
        """
        # pylint:disable=too-many-arguments,too-many-positional-arguments
        LOGGER.debug('query: %s  category: %s  recurse: %s  future: %s',
                     query, category, recurse, future)

        if not self.index:
            return SearchResults([])

        with self.index.searcher() as searcher:
            parsed = self.query_parser.parse(query)

            if category is not None:
                if str(category):
                    # We're in a category other than root
                    cat_term = whoosh.query.Term("category", str(category))
                    if recurse:
                        cat_term = whoosh.query.Or([cat_term,
                                                    whoosh.query.Prefix("category",
                                                                        f"{str(category)}/")])
                    parsed = whoosh.query.And([parsed, cat_term])
                elif not recurse:
                    # We're in root and not recursing, so we need to match empty
                    parsed = whoosh.query.And([parsed, whoosh.query.Term("category", "")])

            if not future:
                parsed = whoosh.query.And([
                    parsed,
                    whoosh.query.Or([
                        whoosh.query.Term("status", model.PublishStatus.PUBLISHED.value),
                        whoosh.query.DateRange("published", None, datetime.datetime.now()),
                    ])
                ])

            LOGGER.debug('parse result: %s', parsed)
            if page is not None:
                results = searcher.search_page(parsed, page, pagelen=count)
            else:
                results = searcher.search(parsed, limit=count)

            return SearchResults(results)
